.wy-nav-side p.caption {
  color: #5197BA;
}

.wy-nav-side a {
  color: #D6D6D6;
}


div.wy-side-nav-search {
  background: #BF3B46;
}

div.wy-side-nay {
  background: #413F3F;
}

table.field-list td li {
  line-height: 18px;
}

table.docutils td p {
  font-size: 14px !important;
  margin-bottom: 6px;

}

table.docutils td li {
  line-height: 18px;

}

/* override table width restrictions */
@media screen and (min-width: 767px) {

    .wy-table-responsive table td {
        /* !important prevents the common CSS stylesheets from overriding
           this as on RTD they are loaded after this stylesheet */
        white-space: normal !important;
    }
    .wy-table-responsive {
        overflow: visible !important;
    }
}
